// UDPAdvertiser
{: Copyright 2000 - Eric Grange (egrange@infonie.fr)<br>
   Home page : http://glscene.org<br>
   License   : Open-source under MPL (see http://www.mozilla.org/mpl)<p>

	Small set of components to advertise a network service (or server).<p>

   Place an active TUDPAdvertiser on the server, and use TUDPAdUser on the
   client side to detect the advertiser you placed on the server.<br>
   This can be particularly usefull for automatic (or assisted) configuration
   of clients : run an advertiser on the app server, and the AdUser components
   will help you identify it (or them if you have multiple servers).<p>

   The only requirements are :<ul>
      <li>allocating a fixed IP port
      <li>IP broadcasting must be possible between server & clients
   <ul>The network load generated by the broadcast is minute (less than the
   usual DNS/WINS traffic), and only clients initiate a broadcast, without
   clients, the advertiser is idle.<p>

   Note that both component can be run on the same machine simultaneously, but
   only one TUDPAdvertiser will be allowed for any single port.<p> 

	<b>History : </b><font size=-1><ul>
      <li>22/10/00 - EG - Added auto-deactivation for TUDPAdUser
	   <li>21/10/00 - EG - Creation
	</ul></font>
}
unit UDPAdvertiser;

interface

uses Windows, Classes, Winsock;

type

	// TUDPAdvertiser
	//
   {: Run this component on the server to advertise a service or server. }
	TUDPAdvertiser = class (TComponent)
	   private
	      { Private Declarations }
         FActive : Boolean;
         FPort : Integer;
         FDefaultInfoString : String;
         FThread : TThread;

	   protected
	      { Protected Declarations }
         procedure Loaded; override;
         procedure SetPort(const val : Integer);
         procedure SetActive(const val : Boolean);
         procedure StartService;
         procedure StopService;

	   public
	      { Public Declarations }
	      constructor Create(AOwner: TComponent); override;
         destructor Destroy; override;

	   published
	      { Published Declarations }
         {: If true advertising is active.<p>
            This component can (and should) remain active all the time, since
            it induces no network load as long as no client is active. }
         property Active : Boolean read FActive write SetActive default True;
         {: IP Port used for the advertisement. }
         property Port : Integer read FPort write SetPort;
         {: Default content for the info string.<p>
            If left empty, server name is returned. This string should be
            kept really short (less than 500 bytes, but the recommended length
            is less than 128 bytes). }
         property DefaultInfoString : String read FDefaultInfoString write FDefaultInfoString;
	end;

   TAdvertisementReceivedEvent = procedure (sender : TObject;
               const infoString, ipAddress, computerName : String) of object;

	// TDetectedAdvertiser
	//
   {: Stores info on a detected advertiser (see TUDPAdUser). }
	TDetectedAdvertiser = class (TCollectionItem)
	   private
	      { Private Declarations }
         FIPAddress : String;
         FComputerName : String;
         FInfoString : String;

      public
	      { Public Declarations }
	      procedure Assign(Source: TPersistent); override;

         property IPAddress : String read FIPAddress;
         property ComputerName : String read FComputerName;
         property InfoString : String read FInfoString;
	end;

	// TDetectedAdvertisers
	//
   {: Colection of TDetectedAdvertiser items. }
	TDetectedAdvertisers = class (TCollection)
	   protected
	      { Protected Declarations }
	      owner : TComponent;
	      function GetOwner: TPersistent; override;
         procedure SetItems(index : Integer; const val : TDetectedAdvertiser);
	      function GetItems(index : Integer) : TDetectedAdvertiser;
      public
	      { Public Declarations }
	      constructor Create(AOwner : TComponent);
         function Add: TDetectedAdvertiser;
	      property Items[index : Integer] : TDetectedAdvertiser read GetItems write SetItems; default;
         function IndexOfIP(const ipAddress : String) : Integer;
   end;

	// TUDPAdUser
	//
   {: Use this component on the client sized to identify advertised servers. }
	TUDPAdUser = class (TComponent)
	   private
	      { Private Declarations }
         FActive : Boolean;
         FPort : Integer;
         FBroadCastInterval : Integer;
         FOnAdvertisementReceived : TAdvertisementReceivedEvent;
         FThread : TThread;
         FDetectedAdvertisers : TDetectedAdvertisers;
         FOnNewAdvertiserDetected : TNotifyEvent;
         FAutoDeactivationDelay : Integer;

	   protected
	      { Protected Declarations }
         procedure Loaded; override;
         procedure SetPort(const val : Integer);
         procedure SetBroadCastInterval(const interval : Integer);
         procedure SetActive(const val : Boolean);
         procedure StartService;
         procedure StopService;

	   public
	      { Public Declarations }
	      constructor Create(AOwner: TComponent); override;
         destructor Destroy; override;

         {: Lists the advertisers that have been detected.<p>
            The list is automatically cleared when Active goes from False to True. }
         property DetectedAdvertisers : TDetectedAdvertisers read FDetectedAdvertisers;

	   published
	      { Published Declarations }
         {: If true advertising search is on.<p>
            This component shouldn't be left active for long, after a few seconds
            you can be sure of having detected all the "reliable" servers on
            the LAN. }
         property Active : Boolean read FActive write SetActive default True;
         {: Port at which servers will be looked up. }
         property Port : Integer read FPort write FPort;
         {: Interval between broadcast (in msec), when component is active.<p>
            Minimum value is 100 ms, and may only work on very fast LANs,
            default value of 1000 ms should work on most LANs.  }
         property BroadCastInterval : Integer read FBroadCastInterval write SetBroadCastInterval default 1000;
         {: Delay before auto-deactivation (in msec).<p>
            If set to <=0 (zero) auto-deactivation is disabled, if  this value
            is positive the component will automatically deactivate after the
            given number of msec. }
         property AutoDeactivationDelay : Integer read FAutoDeactivationDelay write FAutoDeactivationDelay default 3500;

         {: Triggered whenever a new advertiser is detected.<p>
            The new advertiser is the last entry in DetectedAdvertisers. }
         property OnNewAdvertiserDetected : TNotifyEvent read FOnNewAdvertiserDetected write FOnNewAdvertiserDetected;
         {: Use this event to be informed of detected advertisers.<p>
            Not that this event is triggered for every advertisement received. }
         property OnAdvertisementReceived : TAdvertisementReceivedEvent read FOnAdvertisementReceived write FOnAdvertisementReceived;
	end;

procedure Register;

// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------
implementation
// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------

uses SysUtils;

type

   PCardinal = ^Cardinal;

   // TAdvertisingThread
   //
   TAdvertisingThread = class (TThread)
      FAdvertiser : TUDPAdvertiser;
      FSocket : TSocket;
      FAddr : TSockAddr;
      procedure Execute; override;
   end;

   // TUserThread
   //
   TUserThread = class (TThread)
      FUser : TUDPAdUser;
      FSocket : TSocket;
      FAddr : TSockAddr;
      FAdvertiserInfo : String;
      FAdvertiserAddress : String;
      FAutoDeactivationDelay : Integer;
      procedure Execute; override;
      procedure AdvertisementReceived;
   end;

procedure Register;
begin
   RegisterComponents('UDPAdvertiser', [TUDPAdvertiser, TUDPAdUser]);
end;

// CRCInt
//
function CRCInt(const buf : PChar; n : Integer) : Integer;
var
   s, x : Byte;
begin
   s:=0;
   x:=0;
   // cheap CRC...
   while n>0 do begin
      Dec(n);
      s:=s+Integer(buf[n]);
      x:=x xor Integer(buf[n]);
   end;
   Result:=(n shl 16)+(s shl 8)+x;
end;

// ------------------
// ------------------ TUDPAdvertiser ------------------
// ------------------

// Execute
//
procedure TAdvertisingThread.Execute;
var
   err : Integer;
   size, n : Integer;
   wData : TWSAData;
   origAddr : TSockAddr;
   inBuf : array [0..63] of Char;
   buf, computerNameInfoString : String;
begin
   // initialize WinSock DLL
	err:=WSAStartup($0101, wData);
   Assert((err=0), 'Failed to initialize WinSocket, error #'+IntToStr(err));
   // get an udp socket
   FSocket:=socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
   Assert((FSocket<>INVALID_SOCKET), 'Failed to create a socket, Error #'+IntToStr(WSAGetLastError));
   try
      // bind it to required port
      FAddr.sin_family:=AF_INET;
      FAddr.sin_port:=htons(FAdvertiser.FPort);
      FAddr.sin_addr.s_addr:=INADDR_ANY;
      if bind(FSocket, FAddr, SizeOf(FAddr))=SOCKET_ERROR then
         Assert(False, 'Failed to bind a socket, Error #'+IntToStr(WSAGetLastError));
      // non-blocking socket, please
      size:=-1;
      ioctlsocket(FSocket, FIONBIO, size);
      // prepare computer name info string
      n:=MAX_COMPUTERNAME_LENGTH+1;
      SetLength(computerNameInfoString, MAX_COMPUTERNAME_LENGTH+1);
      GetComputerName(PChar(computerNameInfoString), PCardinal(@n)^);
      SetLength(computerNameInfoString, n);
      // infernal loop begins here
      size:=SizeOf(origAddr);
      while not Terminated do begin
         n:=recvfrom(FSocket, inBuf[0], 64, 0, origAddr, size);
         if n>0 then begin
            // incoming ! send back info string
            buf:='abcd'+computerNameInfoString+','+FAdvertiser.FDefaultInfoString;
            PInteger(@buf[1])^:=CRCInt(@buf[5], Length(buf)-4);
            sendto(FSocket, buf[1], Length(buf), 0, origAddr, size);
         end else Sleep(1);
      end;
   finally
      // release the socket
      closesocket(FSocket);
   end;
end;

// ------------------
// ------------------ TUserThread ------------------
// ------------------

// Execute
//
procedure TUserThread.Execute;
var
   err : Integer;
   size, n : Integer;
   wData : TWSAData;
   origAddr : TSockAddr;
   destAddr : TSockAddr;
   inBuf : array [0..511] of Char;
   outBuf : String;
   lastBroadCastTime, deactivationTime, curTime : TDateTime;
begin
   // initialize WinSock DLL
	err:=WSAStartup($0101, wData);
   Assert((err=0), 'Failed to initialize WinSocket, error #'+IntToStr(err));
   // get an udp socket
   FSocket:=socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
   Assert((FSocket<>INVALID_SOCKET), 'Failed to create a socket, Error #'+IntToStr(WSAGetLastError));
   try
      // bind it to an available port
      FAddr.sin_family:=AF_INET;
	   FAddr.sin_port := htons(0);
      FAddr.sin_addr.s_addr:=INADDR_ANY;
      if bind(FSocket, FAddr, SizeOf(FAddr))=SOCKET_ERROR then
         Assert(False, 'Failed to bind a socket, Error #'+IntToStr(WSAGetLastError));
      // non-blocking socket, please
      size:=-1;
      ioctlsocket(FSocket, FIONBIO, size);
      // allow broadcasts
      setsockopt(FSocket, SOL_SOCKET, SO_BROADCAST, PChar(@size), 4);
      // prepare broadcast info & buffers
	   destAddr.sin_family:=AF_INET;
   	destAddr.sin_port:=htons(FUser.FPort);
	   destAddr.sin_addr.s_addr:=Integer(INADDR_BROADCAST);
      outBuf:='blah';
      // deactivation delay
      if FAutoDeactivationDelay>0 then
         deactivationTime:=Now+FAutoDeactivationDelay/(24*3600*1000)
      else deactivationTime:=Now+365*1000; // one thousand years
      // infernal loop begins here
      size:=SizeOf(origAddr);
      lastBroadCastTime:=0;
      while not Terminated do begin
         curTime:=Now;
         if curTime>=deactivationTime then Break;
         if (curTime-lastBroadCastTime)*24*3600*1000>FUser.FBroadCastInterval then begin
            winsock.sendto(FSocket, outBuf[1], Length(outBuf), 0, destAddr, size);
            lastBroadCastTime:=curTime;
         end;
         n:=recvfrom(FSocket, inBuf[0], 512, 0, origAddr, size)-4;
         if n>0 then begin
            // check if packet is valid
            if CRCInt(@inBuf[4], n)=PInteger(@inBuf[0])^ then begin
               SetLength(FAdvertiserInfo, n);
               Move(inBuf[4], FAdvertiserInfo[1], n);
               FAdvertiserAddress:=inet_ntoa(origAddr.sin_addr);
               if not Terminated then
                  Synchronize(AdvertisementReceived);
            end;
         end else Sleep(1);
      end;
   finally
      // release the socket
      closesocket(FSocket);
      FUser.FActive:=False;
      FUser.FThread:=nil;
   end;
end;

// AdvertisementReceived
//
procedure TUserThread.AdvertisementReceived;
var
   i, p : Integer;
   info, compName : String;
begin
   p:=Pos(',', FAdvertiserInfo);
   compName:=Copy(FAdvertiserInfo, 1, p-1);
   info:=Copy(FAdvertiserInfo, p+1, MaxInt);
   i:=FUser.DetectedAdvertisers.IndexOfIP(FAdvertiserAddress);
   if i<0 then begin
      with FUser.DetectedAdvertisers.Add do begin
         FIPAddress:=FAdvertiserAddress;
         FComputerName:=compName;
         FInfoString:=info;
      end;
      if Assigned(FUser.FOnNewAdvertiserDetected) then
         FUser.FOnNewAdvertiserDetected(FUser);
   end else begin
      with FUser.DetectedAdvertisers[i] do begin
         FComputerName:=compName;
         FInfoString:=info;
      end;
   end;
   if Assigned(FUser.FOnAdvertisementReceived) then
      FUser.FOnAdvertisementReceived(FUser, info, FAdvertiserAddress, compName);
end;

// ------------------
// ------------------ TUDPAdvertiser ------------------
// ------------------

// Create
//
constructor TUDPAdvertiser.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
   FActive:=True;
end;

// Destrot
//
destructor TUDPAdvertiser.Destroy;
begin
   Active:=False;
	inherited Destroy;
end;

// Loaded
//
procedure TUDPAdvertiser.Loaded;
begin
   inherited;
   if Active then StartService;
end;

// SetPort
//
procedure TUDPAdvertiser.SetPort(const val : Integer);
begin
   StopService;
   FPort:=val;
   StartService;
end;

// SetActive
//
procedure TUDPAdvertiser.SetActive(const val : Boolean);
begin
   if val<>FActive then begin
      if val then
         StartService
      else StopService;
   end;
end;

// StartService
//
procedure TUDPAdvertiser.StartService;
begin
   if Active then StopService;
   if [csLoading, csDesigning]*ComponentState=[] then begin
      FThread:=TAdvertisingThread.Create(True);
      with TAdvertisingThread(FThread) do begin
         FAdvertiser:=Self;
         FreeOnTerminate:=True;
         Resume;
      end;
   end;
   FActive:=True;
end;

// StopService
//
procedure TUDPAdvertiser.StopService;
begin
   if Active then begin
      if Assigned(FThread) then begin
         FThread.Terminate;
         FThread.WaitFor;
         FThread:=nil;
      end;
      FActive:=False;
   end;
end;

// ------------------
// ------------------ TDetectedAdvertiser ------------------
// ------------------

// Assign
//
procedure TDetectedAdvertiser.Assign(Source: TPersistent);
begin
	if Source is TDetectedAdvertiser then begin
      FIPAddress:=TDetectedAdvertiser(Source).FIPAddress;
      FInfoString:=TDetectedAdvertiser(Source).FInfoString;
      FComputerName:=TDetectedAdvertiser(Source).FComputerName;
	end;
	inherited Destroy;
end;

// ------------------
// ------------------ TDetectedAdvertisers ------------------
// ------------------

// Create
//
constructor TDetectedAdvertisers.Create(AOwner : TComponent);
begin
	Owner:=AOwner;
	inherited Create(TDetectedAdvertiser);
end;

// GetOwner
//
function TDetectedAdvertisers.GetOwner: TPersistent;
begin
	Result:=Owner;
end;

// SetItems
//
procedure TDetectedAdvertisers.SetItems(index : Integer; const val : TDetectedAdvertiser);
begin
	inherited Items[index]:=val;
end;

// GetItems
//
function TDetectedAdvertisers.GetItems(index : Integer) : TDetectedAdvertiser;
begin
	Result:=TDetectedAdvertiser(inherited Items[index]);
end;

// Add
//
function TDetectedAdvertisers.Add: TDetectedAdvertiser;
begin
	Result:=(inherited Add) as TDetectedAdvertiser;
end;

// IndexOfIP
//
function TDetectedAdvertisers.IndexOfIP(const ipAddress : String) : Integer;
var
   i : Integer;
begin
   Result:=-1;
   for i:=0 to Count-1 do if Items[i].IPAddress=ipAddress then begin
      Result:=i;
      Break;
   end;
end;

// ------------------
// ------------------ TUDPAdUser ------------------
// ------------------

// Create
//
constructor TUDPAdUser.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
   FBroadCastInterval:=1000;
   FAutoDeactivationDelay:=3500;
   FDetectedAdvertisers:=TDetectedAdvertisers.Create(Self);
end;

// Destroy
//
destructor TUDPAdUser.Destroy;
begin
   Active:=False;
   FDetectedAdvertisers.Free;
	inherited Destroy;
end;

// Loaded
//
procedure TUDPAdUser.Loaded;
begin
   inherited;
   if Active then StartService;
end;

// SetPort
//
procedure TUDPAdUser.SetPort(const val : Integer);
begin
   StopService;
   FPort:=val;
   StartService;
end;

// SetBroadCastInterval
//
procedure TUDPAdUser.SetBroadCastInterval(const interval : Integer);
begin
   if interval<100 then
      FBroadCastInterval:=100
   else FBroadCastInterval:=interval;
end;

// SetActive
//
procedure TUDPAdUser.SetActive(const val : Boolean);
begin
   if val<>FActive then begin
      if val then
         StartService
      else StopService;
   end;
end;

// StartService
//
procedure TUDPAdUser.StartService;
begin
   if Active then StopService;
   FDetectedAdvertisers.Clear;
   if [csLoading, csDesigning]*ComponentState=[] then begin
      FThread:=TUserThread.Create(True);
      with TUserThread(FThread) do begin
         FUser:=Self;
         FAutoDeactivationDelay:=Self.AutoDeactivationDelay;
         FreeOnTerminate:=True;
         Resume;
      end;
   end;
   FActive:=True;
end;

// StopService
//
procedure TUDPAdUser.StopService;
begin
   if Active then begin
      if Assigned(FThread) then begin
         FThread.Terminate;
         FThread.WaitFor;
         FThread:=nil;
      end;
      FActive:=False;
   end;
end;

end.

